{ ******************************************************************
  Definition of error models for curve fitting
  ****************************************************************** }

unit uerrmod;

interface

uses
  utypes, umath;

type
  TErrModel =
  (ERR_CONST,
   ERR_LIN,
   ERR_POL2,
   ERR_POL3,
   ERR_EXPO,
   ERR_POWER);

function ErrFuncName(ErrModel : TErrModel; ErrInY : Boolean) : String;
{ ------------------------------------------------------------------
  Returns the name of the error function
  ErrInY indicates if the error is related to Y or X
  ------------------------------------------------------------------ }

function LastErrParam(ErrModel : TErrModel) : Integer;
{ ------------------------------------------------------------------
  Returns the index of the last error parameter
  ------------------------------------------------------------------ }

function ErrFunc(ErrModel : TErrModel; Z : Float; Theta : TVector) : Float;
{ ------------------------------------------------------------------
  Computes the error (SD) of an observation Y.

  Z may be X or Y. The parameters are Theta[1], Theta[2],...

  The true SD is Theta[0] * ErrFunc, where Theta[0] (equal to the
  residual standard deviation Sr) is estimated by the program.
  ------------------------------------------------------------------ }

implementation

function ErrFuncName(ErrModel : TErrModel; ErrInY : Boolean) : String;
var
  S : string;
begin
  if ErrInY then S := 'y' else S := 'x';
  case ErrModel of
    ERR_CONST : ErrFuncName := 'v = e0';
    ERR_LIN   : ErrFuncName := 'v = e0.(1 + e1.' + S + ')';
    ERR_POL2  : ErrFuncName := 'v = e0.(1 + e1.' + S + '+ e2.' + S + '^2)';
    ERR_POL3  : ErrFuncName := 'v = e0.(1 + e1.' + S + '+ e2.' + S + '^2 + e3.' + S + '^3)';
    ERR_EXPO  : ErrFuncName := 'v = e0.exp(e1.' + S + ')';
    ERR_POWER : ErrFuncName := 'v = e0.' + S + '^e1';
  end;
end;

function LastErrParam(ErrModel : TErrModel) : Integer;
begin
  case ErrModel of
    ERR_CONST : LastErrParam := 0;
    ERR_LIN   : LastErrParam := 1;
    ERR_POL2  : LastErrParam := 2;
    ERR_POL3  : LastErrParam := 3;
    ERR_EXPO  : LastErrParam := 1;
    ERR_POWER : LastErrParam := 1;
  end;
end;

function ErrFunc(ErrModel : TErrModel; Z : Float; Theta : TVector) : Float;
begin
  case ErrModel of
    ERR_CONST : ErrFunc := 1.0;
    ERR_LIN   : ErrFunc := 1.0 + Theta[1] * Z;
    ERR_POL2  : ErrFunc := 1.0 + Z * (Theta[1] + Theta[2] * Z);
    ERR_POL3  : ErrFunc := 1.0 + Z * (Theta[1] + Z * (Theta[2] + Theta[3] * Z));
    ERR_EXPO  : ErrFunc := Exp(Theta[1] * Z);
    ERR_POWER : ErrFunc := Power(Z, Theta[1]);
  end;
end;

end.
