{ ******************************************************************
  This unit fits the reparameterized gamma variate regression model:

             y = ymax * w^alpha * exp[ alpha * (1 - w)]
              
  with w = (x - x0) / (xmax - x0)
  
  Ref.: M. T. Madsen, Phys. Med. Biol. 1992, 37, 1597-1600
  ****************************************************************** }

unit ugamrfit;

interface

uses
  utypes, umath, ulinfit, unlfit;

procedure GammaFit(X, Y    : TVector;
                   Lb, Ub  : Integer;
                   MaxIter : Integer;
                   Tol     : Float;
                   B       : TVector;
                   V       : TMatrix);
{ ------------------------------------------------------------------
  Unweighted fit of model
  ------------------------------------------------------------------
  Input parameters:  X, Y    = point coordinates
                     Lb, Ub  = array bounds
                     MaxIter = max. number of iterations
                     Tol     = tolerance on parameters
  Output parameters: B       = regression parameters
                     V       = inverse matrix
  ------------------------------------------------------------------ }

procedure WGammaFit(X, Y, S : TVector;
                    Lb, Ub  : Integer;
                    MaxIter : Integer;
                    Tol     : Float;
                    B       : TVector;
                    V       : TMatrix);
{ ------------------------------------------------------------------
  Weighted fit of model
  ------------------------------------------------------------------
  Additional input parameter:
  S = standard deviations of observations
  ------------------------------------------------------------------ }

function GammaFit_Func(X : Float; B : TVector) : Float;
{ ------------------------------------------------------------------
  Returns the value of the regression function at point X
  ------------------------------------------------------------------ }

implementation

const
  FirstParam = 0;
  LastParam  = 3;

function GammaFit_Func(X : Float; B : TVector) : Float;
{ ------------------------------------------------------------------
  Computes the regression function at point X
  B is the vector of parameters, such that :

  B[0] = x0     B[1] = xmax     B[2] = ymax     B[3] = alpha
  ------------------------------------------------------------------ }
var
  W, Z : Float;
begin
  if X <= B[0] then
    begin
      GammaFit_Func := 0.0;
      Exit;
    end;

  W := (X - B[0]) / (B[1] - B[0]);  { (x - x0) / (xmax - x0) }
  Z := B[3] * (1.0 + Ln(W) - W);    { alpha * (1 + ln(w) - w }

  if Z > MinLog then
    GammaFit_Func := B[2] * Exp(Z)  { ymax * w^alpha * exp[alpha * (1 - w)] }
  else
    GammaFit_Func := 0.0;
end;

procedure GammaFit_Deriv(X, Y : Float; B, D : TVector);
{ ------------------------------------------------------------------
  Computes the derivatives of the regression function at point X
  with respect to the parameters B. The results are returned in D.
  D[I] contains the derivative with respect to the I-th parameter.
  ------------------------------------------------------------------ }
var
  U, W, Z : Float;
  I       : Integer;
begin
  if X <= B[0] then
    begin
      for I := FirstParam to LastParam do D[I] := 0.0;
      Exit;
    end;

  U := B[1] - B[0];
  W := (X - B[0]) / U;
  Z := B[3] * Y * (1.0 / W - 1.0) / Sqr(U);
  
  D[0] := Z * (X - B[1]);         { dy / d x0 }
  D[1] := Z * (B[0] - X);         { dy / d xmax }
  D[2] := Y / B[2] ;              { dy / d ymax }
  D[3] := Y * (1.0 + Ln(W) - W);  { dy / d alpha }
end;

procedure ApproxFit(Mode    : TRegMode;
                    X, Y, S : TVector;
                    Lb, Ub  : Integer;
                    B       : TVector);
{ ------------------------------------------------------------------
  Computes initial estimates of the regression parameters on the
  linearized form of the function:

  ln y = ln ymax + alpha * (1 + ln w - w), w = (x - x0) / (xmax - x0)
  ------------------------------------------------------------------
  Input :  Mode   = OLS or WLS
           X, Y   = point coordinates
           Lb, Ub = array bounds
  Output : B      = estimated regression parameters
  ------------------------------------------------------------------ }
var
  K  : Integer;
  W  : Float;    
  X1 : TVector;  { 1 + ln w - w }
  Y1 : TVector;  { ln y }
  S1 : TVector;  { Standard deviations of ln y }
  A  : TVector;  { Linear regression parameters }
  P  : Integer;  { Nb of points for linear regression }
  V  : TMatrix;  { Variance-covariance matrix }

begin
  DimVector(X1, Ub);
  DimVector(Y1, Ub);
  DimVector(S1, Ub);
  DimVector(A, 1);
  DimMatrix(V, 1, 1);

  { Find the lowest X having a positive Y }
  W := MaxNum;
  for K := Lb to Ub do
    if (X[K] > 0.0) and (Y[K] > 0.0) and (X[K] < W) then
      W := X[K];

  { Set x0 slightly below this X value }
  B[0] := 0.9 * W;
  
  { Find xmax }
  W := 0.0;
  for K := Lb to Ub do
    if Y[K] > W then
      begin
        B[1] := X[K];  { xmax }
        W := Y[K];     { ymax }
      end;  

  P := Pred(Lb);
  for K := Lb to Ub do
    if Y[K] > 0.0 then
      begin
        Inc(P);
        W := (X[K] - B[0]) / (B[1] - B[0]);
        X1[P] := 1.0 + Ln(W) - W;
        Y1[P] := Ln(Y[K]);
        S1[P] := 1.0 / Y[K];
        if Mode = WLS then S1[P] := S1[P] * S[K];
      end;

  WLinFit(X1, Y1, S1, Lb, P, A, V);

  if MathErr = MatOk then
    begin
      B[2] := Exp(A[0]);  { ymax }
      B[3] := A[1];       { alpha }
    end;
end;

procedure GenGammaFit(Mode    : TRegMode;
                      X, Y, S : TVector;
                      Lb, Ub  : Integer;
                      MaxIter : Integer;
                      Tol     : Float;
                      B       : TVector;
                      V       : TMatrix);
begin
  if (GetOptAlgo in [NL_MARQ, NL_BFGS, NL_SIMP])
     and NullParam(B, FirstParam, LastParam) then
       ApproxFit(Mode, X, Y, S, Lb, Ub, B);

  if MaxIter = 0 then Exit;

  case Mode of
    OLS : NLFit(GammaFit_Func, GammaFit_Deriv, X, Y, Lb, Ub,
                       MaxIter, Tol, B, FirstParam, LastParam, V);
    WLS : WNLFit(GammaFit_Func, GammaFit_Deriv, X, Y, S, Lb, Ub,
                       MaxIter, Tol, B, FirstParam, LastParam, V);
  end;
end;

procedure GammaFit(X, Y    : TVector;
                   Lb, Ub  : Integer;
                   MaxIter : Integer;
                   Tol     : Float;
                   B       : TVector;
                   V       : TMatrix);
begin
  GenGammaFit(OLS, X, Y, nil, Lb, Ub, MaxIter, Tol, B, V);
end;

procedure WGammaFit(X, Y, S : TVector;
                    Lb, Ub  : Integer;
                    MaxIter : Integer;
                    Tol     : Float;
                    B       : TVector;
                    V       : TMatrix);
begin
  GenGammaFit(WLS, X, Y, S, Lb, Ub, MaxIter, Tol, B, V);
end;

end.
